@@ -33,7 +33,7 @@
 
 #include "kip.h"
 
-RCSID("$Id: kipd.c,v 1.22 2001/09/17 04:42:48 assar Exp $");
+RCSID("$Id: kipd.c,v 1.23 2002/09/09 15:57:23 joda Exp $");
 
 static int
 recv_conn (int sock, des_cblock *key, des_key_schedule schedule,
@@ -163,7 +163,7 @@ main (int argc, char **argv)
     int optind = 0;
 
     setprogname (argv[0]);
-    roken_openlog(getprogname(), LOG_PID|LOG_CONS, LOG_DAEMON);
+    roken_openlog(getprogname(), LOG_PID, LOG_DAEMON);
 
     if (getarg (args, sizeof(args) / sizeof(args[0]), argc, argv,
 		&optind))
