@@ -5,7 +5,7 @@
  */
 
 #include <popper.h>
-RCSID("$Id: pop_user.c,v 1.13 1997/10/29 02:47:58 assar Exp $");
+RCSID("$Id: pop_user.c,v 1.15 1999/09/16 20:38:50 assar Exp $");
 
 /* 
  *  user:   Prompt for the user name at the start of a POP session
@@ -16,7 +16,7 @@ pop_user (POP *p)
 {
     char ss[256];
 
-    strcpy(p->user, p->pop_parm[1]);
+    strlcpy(p->user, p->pop_parm[1], sizeof(p->user));
 
 #ifdef OTP
     if (otp_challenge (&p->otp_ctx, p->user, ss, sizeof(ss)) == 0) {
